#!/bin/python

from netCDF4 import Dataset as DS
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from math import *
from numpy import log, vectorize, exp, arange
import sys
from scipy.stats import linregress
from re import split
from os.path import splitext, isfile, dirname, basename

import re

from extrema import *

figure_size = (8, 4)
figure_sizeb = (4, 3)

amu = 1.66053904e-27
q_e = 1.6021766208E-19
m_e = 9.10938356E-31
m_p = 1.672621898e-27
k_B = 1.38064852E-23

time_variable = '$t\\;\\frac{v_T}{\lambda_D}$'

interpolation_val = None  # 'nearest'
interpolation_val = "gaussian"  # 'nearest'

## global variables

def get_markings(name):
    from re import match
    if (name=='heat_flux'):
        return '$q_z$', 'heat flux evolution at the right wall'
    elif (name=='energy_flux'):
        return '$Q_\\parallel/(m v_T^2)$','parallel energy flux at the right wall'
    elif (name=='n'):
        return '$N/N_0$','particle number'
    elif (match(r'.*sum_momentum_flux_l.*',name)):
        return '$N/N_0$','particle loss ($x=x_{{min}}$)'
    elif (match(r'.*sum_momentum_flux_r.*',name)):
        return '$N/N_0$','particle loss ($x=x_{{max}}$)'
    elif (match(r'.*momentum_flux_l.*',name)):
        return '$j_\\parallel$','$\\parallel$ particle flux ($x=x_{{min}}$)'
    elif (match(r'.*momentum_flux_r.*',name)):
        return '$j_\\parallel$','$\\parallel$ particle flux ($x=x_{{max}}$)'
    elif (name=='temperature_p'):
        return '$T_\\parallel$','parallel Temperature in the middle'
    elif (name=='temperature_n'):
        return '$T_\\perp$','perpendicular Temperature in the middle'
    else:
        return '$??$',name+' evolution'

def get_label_name(name):
    from re import sub,match
    if match(r'.*_nu[0-9].*',name):
        nu = sub(r".*_nu([0-9.]*)_.*",r'\1',name)
        if float(nu)<=0.012: nu="0.0"
        #nu = float(nu)
        return '$\\nu={}$'.format(nu)

class globals:
    def __init__(self):
        self.t = {}
        self.data = {}

    def write_evolution_proto(self,file, species, data, target):
        if species not in target:
            #import pdb; pdb.set_trace()
            target[species]={file: data}
        else:
            target[species][file]=data

    def write_evolution(self,file, species, type, data):
        if type=="t":
            if file not in self.t: self.t[file]=data
        else:
            if type not in self.data:
                self.data[type] = {}
            self.write_evolution_proto(file,species,data,self.data[type])

    def plot_specific(self,target,species):

        print('plotting',target,species)
        ylabel,label = get_markings(target)
        plt.figure(figsize=figure_size)
        #import pdb; pdb.set_trace()
        base = [ x for x,y in self.data[target][species].items()][0]
        base = dirname(base)+'/'

        for key, val in self.data[target][species].items():
            pl = plt.plot(self.t[key], val,label=get_label_name(basename(key)))
            if target in ['n','momentum_flux_r', 'momentum_flux_l', 'energy_flux', 'heat_flux']:
                #xm,ym=find_global_maximum(self.t[key],val)
                xtr = extrema(val,self.t[key])
                #if xtr.maximum['x']>0.1*self.t[key][-1] and not isnan(xtr.maximum['y']):
                    ##import pdb; pdb.set_trace()
                    ##print("here",xm,ym)
                    #try:
                        #plt.text(xtr.maximum['x'],xtr.maximum['y'],"$({:.2f},{:.2f})$".format(xtr.maximum['x'],xtr.maximum['y']), fontsize=6, color=pl[0].get_color(),alpha=0.5)
                    #except:
                        #import pdb; pdb.set_trace()

                #plt.axvline(x=self.t[key][-1],color=pl[0].get_color(),alpha=.1,linestyle='-.')

        plt.xlabel(time_variable)
        plt.ylabel(ylabel)
        plt.tight_layout()
        plt.legend(loc='best')
        plt.savefig((base+species+'_'+target+'_evolution.pdf'), transparent=True, format="pdf", bbox_inches = 'tight')
        plt.close()

    def plot(self):
        #import pdb; pdb.set_trace()
        for target,species in self.data.items():
            for sp,data in species.items():
                self.plot_specific(target,sp)


gl_vars =  globals()


def num_to_str(num):
    if (abs(num) < .01 or abs(num) > 999) and abs(num) != 0.:
        from re import sub
        output = '{:.2e}'.format(num)
        output = sub(r'(-?\d*\.?\d?)e\+?(-?)0*(\d*)', r'\1{\\cdot}10^{{\2\3}}', output)
    else:
        output = '{:.2}'.format(num)
    return output

def set_im_ticks(cb,n0=1.):
    try:
        #cb.set_ticks(cb.get_ticks())
        cb.ax.set_yticklabels(['$' + num_to_str(x) + '$' for x in cb.ax.get_yticks() / n0])
    except:
        import pdb; pdb.set_trace()

def v_thermal_K(temp_in_k, mass=m_e):
    return sqrt(2 / mass * temp_in_k * k_B)


def maxwellian(v, v_thermal, mu=0.):
    return exp(-((v - mu) / v_thermal)**2) / (sqrt(pi) * v_thermal)


vmaxwellian = vectorize(maxwellian)


def get_m_of_n(liste, m):
    from math import ceil

    pos = []
    val = []

    shift = (len(liste) - 1) / (m - 1)

    for i in range(m):
        pos.append(int(ceil(i * shift)))
        val.append(round(liste[pos[-1]], 2))

    return pos, val


# making cuts
print("making cuts")

def find_local_maxima(a,x_data=None):
    output_x=[]
    output_y=[]

    #import pdb; pdb.set_trace()
    for i in range(len(a))[1:-1]:
        if a[i] > a[i-1] and a[i]>a[i+1]:
            output_y.append(a[i])
            if x_data is None:
                output_x.append(i)
            else:
                output_x.append(x_data[i])

    return output_x,output_y

def find_global_maximum(x,y):
    aval=x[0]
    bval=y[0]
    for a,b in zip(x,y):
        if b>bval:
            bval=b
            aval=a
    return aval,bval

def find_global_minimum(x,y):
    aval=x[0]
    bval=y[0]
    for a,b in zip(x,y):
        if b<bval:
            bval=b
            aval=a
    return aval,bval

def integrate_flux(vals,dt):
    output = [vals[0]/2.*dt]
    for a,b in zip(vals[0:-1],vals[1:]):
        output.append(output[-1]+(a+b)/2.*dt)

    return output


class plot_nc_file:
    def __init__(self,fl,draw_evolutions=True):
        self.fl = fl
        self.ds = DS(fl)


        print("Plotting",fl)
        self.plot_electric_energy()
        self.plot_e_field()

        for key,val in self.ds.groups.items():
            self.plot_species(val,self,draw_evolutions)

    class plot_species:
        def __init__(self,grp,parent,draw_evolutions):
            self.group = grp
            self.parent = parent
            self.description = grp['description'][0]
            self.base_name = self.parent.fl+'_'+self.description.split(' ')[0]
            #print('basename:',self.base_name,self.description.split(' '))

            self.species = ''
            self.init_txt()

            self.vp_plot()
            self.vn_plot()
            self.vp_cut()
            self.vn_cut()
            if draw_evolutions:
                self.plot_flux2(['n'],['momentum_flux_l', 'momentum_flux_r'])
                #self.plot_flux('n')
                self.plot_flux('heat_flux')
                self.plot_flux('momentum_flux_l')
                self.plot_flux('momentum_flux_r')
                self.plot_flux('energy_flux')
                self.plot_flux('temperature_p')
                self.plot_flux('temperature_n')

        def vp_plot(self):
            min_vp = 0
            max_vp = -1
            #min_vp = floor(ds['f'].shape[2]/4.)
            #max_vp = ceil(3*ds['f'].shape[2]/4.)
            try:
                plt.figure(figsize=figure_sizeb)
                pos = 0 #ceil(self.group['f'].shape[1] / 2)
                pos = ceil(self.group['f'].shape[1] / 5/sqrt(2))
                txt2 = '$v_{{\\perp}}/v_T={}$\n'.format(num_to_str(self.group['vn'][pos].data))
                plt.imshow(self.group['f'][:, pos, min_vp:max_vp].T, aspect='auto', interpolation=interpolation_val, vmin=0.)
                # plt.set_cmap('ocean')
                pos, val = get_m_of_n(self.parent.ds['x'][:], 5)
                plt.xticks(pos, val)
                plt.xlabel('$x/\\lambda_D$')
                pos, val = get_m_of_n(self.group['vp'][min_vp:max_vp], 5)
                val = [round(float(x), 1) for x in val]
                plt.yticks(pos, val)
                plt.ylabel('$v_\\parallel/v_T$')
                #import pdb; pdb.set_trace()

                plt.title(self.description)
                plt.text(1, 1, self.txt + txt2, color='white', fontsize=8, verticalalignment='top')
                cb = plt.colorbar()
                set_im_ticks(cb,self.group['n0'][0].data)
                #cb.set_ticks(cb.get_ticks())
                #cb.set_ticklabels(['$' + num_to_str(x) + '$' for x in cb.get_ticks() / self.group['n0'][0].data])
                cb.set_label("$f/n_0$")
                # cb.title('$f$')
                plt.tight_layout()
                plt.savefig(self.base_name + '_vp.pdf', transparent=True, format="pdf", bbox_inches = 'tight')
            except:
                import pdb; pdb.set_trace()
            plt.close()

        def vn_plot(self):
            try:
                plt.figure(figsize=figure_sizeb)
                pos = ceil(self.group['f'].shape[2] / 2)
                txt2 = '$v_{{\\parallel}}/v_T={}$\n'.format(num_to_str(self.group['vp'][pos].data))
                plt.imshow(self.group['f'][:, :, pos].T, aspect='auto', interpolation=interpolation_val)
                pos, val = get_m_of_n(self.parent.ds['x'][:], 5)
                plt.xticks(pos, val)
                plt.xlabel('$x/\\lambda_D$')
                pos, val = get_m_of_n(self.group['vn'][:], 5)
                val = [round(float(x), 1) for x in val]
                plt.yticks(pos, val)
                plt.ylabel('$v_\\perp/v_T$')

                plt.title(self.description)
                plt.text(1, 1, self.txt + txt2, color='white', fontsize=10, verticalalignment='top')
                cb = plt.colorbar()
                cb.set_label("$f/n_0$")
                set_im_ticks(cb,self.group['n0'][0].data)
                #cb.set_ticks(cb.get_ticks())
                #cb.set_ticklabels(['$' + num_to_str(x) + '$' for x in cb.get_ticks() / self.group['n0'][0].data])
                plt.tight_layout()
                plt.savefig(self.base_name + '_vn.pdf', transparent=True, format="pdf", bbox_inches = 'tight')
            except:
                import pdb; pdb.set_trace()
            plt.close()

        def vp_cut(self, pos=None):
            try:
                #maxw=vmaxwellian(ds['v'][:],1)
                #import pdb; pdb.set_trace()
                if pos == None:
                    pos = ceil(self.group['f'].shape[1] / 2)
                txt2 = '$v_{{\\perp}}/v_T={}$\n'.format(num_to_str(self.group['vn'][pos].data))
                plt.figure(figsize=figure_size)
                x = self.group['vp'][:]
                plt.plot(x, self.group['f'][0, pos, :] / self.group['n0'][0].data, linestyle='-.', label="x=0%")
                plt.plot(x, self.group['f'][int(self.parent.ds['x'].size / 4), pos, :] / self.group['n0'][0].data, linestyle='--', label="x=25%")
                plt.plot(x, self.group['f'][int(self.parent.ds['x'].size / 2), pos, :] / self.group['n0'][0].data, linestyle='--', label="x=50%")
                plt.plot(x, self.group['f'][int(self.parent.ds['x'].size / 4 * 3), pos, :] / self.group['n0'][0].data, linestyle='--', label="x=75%")
                plt.plot(x, self.group['f'][-1, pos, :] / self.group['n0'][0].data, linestyle='--', label="x=100%")
                plt.legend(loc="best")
                plt.figtext(0.15, 0.88, self.txt + txt2, color='k', fontsize=10, verticalalignment='top')
                plt.ylabel("$f$")
                plt.xlabel('$v_\\parallel/v_T$')
                #plt.plot(ds['v'],vmaxwellian(ds['v'][:],1))
                plt.title(self.description)
                plt.tight_layout()
                plt.savefig(self.base_name + '_pcut.pdf', transparent=True, format="pdf", bbox_inches = 'tight')
            except:
                import pdb; pdb.set_trace()
            plt.close()

        def vn_cut(self, pos=None):
            try:
                #maxw=vmaxwellian(ds['v'][:],1)
                #import pdb; pdb.set_trace()
                if pos == None:
                    pos = ceil(self.group['f'].shape[2] / 2)
                txt2 = '$v_{{\\parallel}}/v_T={}$\n'.format(num_to_str(self.group['vp'][pos].data))
                plt.figure(figsize=figure_size)
                x = self.group['vn'][:]
                plt.plot(x, self.group['f'][0, :, pos] / self.group['n0'][0].data, linestyle='-.', label="x=0%")
                plt.plot(x, self.group['f'][int(self.parent.ds['x'].size / 4), :, pos] / self.group['n0'][0].data, linestyle='--', label="x=25%")
                plt.plot(x, self.group['f'][int(self.parent.ds['x'].size / 2), :, pos] / self.group['n0'][0].data, linestyle='--', label="x=50%")
                plt.plot(x, self.group['f'][int(self.parent.ds['x'].size / 4 * 3), :, pos] / self.group['n0'][0].data, linestyle='--', label="x=75%")
                plt.plot(x, self.group['f'][-1, :, pos] / self.group['n0'][0].data, linestyle='--', label="x=100%")
                plt.legend(loc="best")
                plt.figtext(0.15, 0.88, self.txt + txt2, color='k', fontsize=10, verticalalignment='top')
                plt.ylabel("$f$")
                plt.xlabel('$v_\\perp/v_T$')
                #plt.plot(ds['v'],vmaxwellian(ds['v'][:],1))
                plt.title(self.description)
                plt.tight_layout()
                plt.savefig(self.base_name + '_ncut.pdf', transparent=True, format="pdf", bbox_inches = 'tight')
            except:
                import pdb; pdb.set_trace()
            plt.close()

        def init_txt(self):
            #import pdb; pdb.set_trace()
            t_string = "$t={}$".format(self.parent.ds['electric_energy'].shape[0]*self.parent.ds['dt'][0].data)


            res_string = "#domain={{{},{},{}}}".format(self.group['f'].shape[0],self.group['f'].shape[1],self.group['f'].shape[2])

            m = self.group['m'][0].data
            assert m > 0.
            if m<.9*amu:
                m = m / m_e
                assert m % 1 < 1e-3
                m_string = "$m={:d} m_e$".format(int(m))
                self.species = 'electron'
            elif m/amu % 1 <1e-3:
                m = m / amu
                m_string = "$m={:d} m_u$".format(int(m))
                self.species = 'ion'
            elif m/m_p % 1 <1e-3:
                m = m / m_p
                m_string = "$m={:d} m_p$".format(int(m))
                self.species = 'ion'
            else:
                import pdb; pdb.set_trace()
                raise ValueError("Error: mass m is not a valid multiple of amu, m_p or m_e" % m)

            q = self.group['q'][0].data
            # import pdb; pdb.set_trace()
            if abs(q) < 0.95:
                q = q / q_e
            assert abs(q) % 1. < 1e-5
            q=int(q)
            q_string = "$q={:d}e$".format(q, 0)
            # import pdb; pdb.set_trace()

            try:
                self.txt = ""
                #self.txt += t_string + '\n'
                #self.txt += res_string + '\n'
                self.txt += m_string + '\n'
                self.txt += q_string + '\n'
                self.txt += '$v_{{unit}}={}\\,$m/s\n'.format(num_to_str(self.parent.ds['v_normalizaion'][0].data))
                #self.txt += '$n_0={} m^{{-3}}$\n'.format(num_to_str(self.group['n0'][0].data))
                self.txt += '$\lambda={}$\n'.format(num_to_str(self.group['lambda'][0].data))
            except:
                import pdb; pdb.set_trace()



        def plot_flux(self,flux='heat_flux'):
            #import pdb; pdb.set_trace()
            if flux+'_evolution' not in self.group.variables:
                print('Warning:',flux,'has not been found in the given netCDF file.')
                return 0
            if self.group[flux+'_evolution'].size == 0:
                return 0

            ylabel,label = get_markings(flux)

            t = range(self.group[flux+'_evolution'][:].size) * self.parent.ds['dt'][0].data *2 / self.parent.ds['X_length'][0].data
            plt.figure(figsize=figure_size)
            ##plt.plot(ds['electric_energy'][:])

            plt.plot(t, self.group[flux+'_evolution'][:],label=label)
            gl_vars.write_evolution(self.parent.fl,self.species, "t", t)
            gl_vars.write_evolution(self.parent.fl,self.species, flux, self.group[flux+'_evolution'][:])
            plt.xlabel(time_variable)
            plt.ylabel(ylabel)
            plt.tight_layout()
            plt.legend(loc='best')
            plt.savefig((self.base_name + '_'+flux+'.pdf'), transparent=True, format="pdf", bbox_inches = 'tight')
            plt.close()

        def plot_flux2(self,fluxes=['heat_flux'],secondary=[]):
            ##import pdb; pdb.set_trace()
            if isinstance(fluxes,str): fluxes=list(fluxes)

            plt.figure(figsize=figure_size)
            store = []
            normal = None
            for flux in fluxes+secondary:
                if flux+'_evolution' not in self.group.variables:
                    print('Warning:',flux,'has not been found in the given netCDF file.')
                    return 0
                if self.group[flux+'_evolution'].size == 0:
                    return 0


                t = range(self.group[flux+'_evolution'][:].size) * self.parent.ds['dt'][0].data *2 / self.parent.ds['X_length'][0].data
                #import pdb; pdb.set_trace()

                data = self.group[flux+'_evolution'][:]
                if normal is None: normal=data[0]
                if flux in secondary:
                    data = integrate_flux(data,self.parent.ds['dt'][0].data)
                    if len(fluxes)==1:
                        for i in range(len(data)):
                            store[i]+=data[i]
                    ylabel,label = get_markings('sum_'+flux)
                else:
                    if len(fluxes)==1:
                        store=data
                    gl_vars.write_evolution(self.parent.fl,self.species, "t", t)
                    gl_vars.write_evolution(self.parent.fl,self.species, flux, data)
                    ylabel,label = get_markings(flux)
                    #import pdb; pdb.set_trace()
                pl=plt.plot(t, data/normal,label=label)
                if True or flux in ['n','momentum_flux_r', 'momentum_flux_l', 'energy_flux', 'heat_flux']:

                    xtr = extrema(data,t)
                    for k,v in xtr.poi().items():
                        #import pdb; pdb.set_trace()
                        plt.text(v['x'],v['y'],"$({:.2f},{:.4f})$".format(v['x'],v['y']), fontsize=6, color=pl[0].get_color(),alpha=0.75)

            if len(fluxes)==1 and len(secondary)>0:
                pl=plt.plot(t,store/normal,label="total particle number")
                xtr = extrema(store,t)
                for k,v in xtr.poi().items():
                    plt.text(v['x'],v['y'],"$({:.2f},{:.4f})$".format(v['x'],v['y']), fontsize=6, color=pl[0].get_color(),alpha=0.75)

            if len(fluxes)==1:
                gl_vars.write_evolution(self.parent.fl,self.species, "t", t)
                gl_vars.write_evolution(self.parent.fl,self.species, '_'.join(fluxes+secondary), store)


            plt.xlabel(time_variable)
            plt.ylabel(ylabel)
            plt.tight_layout()
            plt.legend(loc='best')
            oname=self.base_name + '_'+'_'.join(fluxes)+'.pdf'
            #print("Printing",oname)
            plt.savefig(oname, transparent=True, format="pdf", bbox_inches = 'tight')
            plt.close()

    def plot_e_field(self):
        plt.figure(figsize=figure_size)
        plt.plot(self.ds['x'], self.ds['E'][:])
        plt.xlabel('$x/\\lambda_D$')
        plt.ylabel('$E$')
        plt.tight_layout()
        plt.savefig((self.fl + '_e_field.pdf'), transparent=True, format="pdf", bbox_inches = 'tight')
        plt.close()

    def plot_electric_energy(self):
        #import pdb; pdb.set_trace()
        if self.ds['electric_energy'].size == 0:
            return 0

        dt = self.ds['dt'][0].data
        t = range(self.ds['electric_energy'][:].size) * dt
        plt.figure(figsize=figure_size)
        ##plt.plot(ds['electric_energy'][:])
        plt.semilogy(t, self.ds['electric_energy'][:],label="Electric energy")

        x,y = find_local_maxima(self.ds['electric_energy'][:],t)

        ii = 0
        #print(x)
        while ii<len(x):
            #print("ii:",ii)
            xx = []
            yy = []
            for jj in range(ii,len(x)):
                if jj==ii or y[jj-1]>=y[jj]:
                    xx.append(x[jj])
                    yy.append(log(y[jj]))
                    if jj == len(x)-1: ii=len(x)
                else:
                    ii=jj
                    break

            #print("pos0",x,y)
            #print("pos",xx,yy)

            if len(xx)>1:
                affine = linregress(xx,yy)
                #print(affine)
                period = 0
                l=0
                for p in [j-i for i, j in zip(xx[:-1], xx[1:])]:
                    period += p
                    l += 1
                period /= l
                #print(affine[0],affine[1],period)
                mn=xx[0]-2.
                if mn<t[0]: mn=t[0]
                mx=xx[-1]+2.
                if mx>t[-1]: mx=t[-1]
                mx+=dt
                tt=arange(mn,mx,dt)
                y2 = exp(affine[1]+affine[0]*tt)
                #print(tt,y2)
                try:
                    plt.semilogy(tt, y2, label="linfit: ${}\cdot e^{{ {:.3f} t}}$\nT={:.2f}, $dt={}$".format(num_to_str(exp(affine[1])),affine[0],period,num_to_str(self.ds['dt'][0].data)))
                except:
                    pass

        plt.xlabel(time_variable)
        plt.ylabel('$W_E$')
        plt.tight_layout()
        plt.legend(loc='best')
        plt.savefig((self.fl + '_el_energy.pdf'), transparent=True, format="pdf", bbox_inches = 'tight')
        plt.close()

def get_head_last(args):
    from re import sub
    tmp =  {}
    for x in args:
        head = sub(r'(.*)_[^_]*\.nc',r'\1',x)
        tail = sub(r'.*_([^_]*)\.nc',r'\1',x)
        if head not in tmp: tmp[head]=[tail]
        else: tmp[head].append(tail)

    #import pdb; pdb.set_trace()
    return [x+'_'+max(y)+'.nc' for x,y in tmp.items()]

if __name__ == "__main__":
    #highest_file :

    ## options
    all = True if "all" in sys.argv[1:] else False
    movie = True if "movie" in sys.argv[1:] else False

    flst=[x for x in sys.argv[1:] if x.endswith('.nc') and isfile(x)]
    #import pdb; pdb.set_trace()
    last = get_head_last(flst)

    # overlay plot_species
    for fle in last:
        plot_nc_file(fle,True)
    if len(last)>0:
        gl_vars.plot()

    if all:
        from glob import glob
        from os.path import getmtime
        for fle in flst:
            pdfs = glob(fle+'*.pdf')
            if len(pdfs)>0:
                if getmtime(fle) < getmtime(pdfs[0]):
                    #print(fle,"older than",pdfs[0])
                    continue
            plot_nc_file(fle,False)

    elif movie:
        pass

