#pragma once

#define _USE_MATH_DEFINES
#include <cmath>
#include <cassert>
#include <array>

// #define pi  M_PI
constexpr double amu = 1.66053904e-27; // kg
constexpr double m_e = 9.10938356E-31; // kg
constexpr double m_p = 1.672621898e-27; // kg
constexpr double k_B = 1.38064852E-23; // J/K
constexpr double landau = 10.;

constexpr double c_l = 299792458; // m/s

constexpr double q_e = 1.6021766208E-19; // C
constexpr double mu_0 = 4*M_PI*1e-7; // H/m
constexpr double epsilon_0 = 1./(mu_0*c_l*c_l); // F/m

template<typename T>
T v_thermal_K(T temp_in_K, T mass = m_e){
    return std::sqrt(2/mass*temp_in_K*k_B);
}
template<typename T>
T v_thermal_eV(T temp_in_eV, T mass = m_e){
    return std::sqrt(2/mass*q_e*temp_in_eV);
}

template<size_t dim=1, typename T=double>
T maxwellian(T v, T v_thermal, T mu=0.){
//     assert(v_thermal>0.);
    static_assert(dim>0);

    if(v_thermal<=0.){
        return 0.;
    }
    T output = std::exp(-std::pow((v-mu)/v_thermal,2));
    output /=std::pow(std::sqrt(M_PI)*v_thermal,dim);
    if constexpr (dim>1) {
        output *= 2*(dim-1)*M_PI;
    }
    output *= std::pow(v,dim-1);
    return output;
}
template<typename T, size_t dim>
T maxwellian(const std::array<T,dim>& v, T v_thermal){
    if(v_thermal<=0.){
        return 0.;
    }
    T tmp_v = 0;
    for(size_t i=0; i<dim; ++i) {
        tmp_v+=std::pow(v[i],2);
    }
    return std::exp(-tmp_v*std::pow(v_thermal,-2))/std::pow(std::sqrt(M_PI)*v_thermal,dim);
}
