#pragma once

// #include <iostream>
#include <netcdf>
#include <string>

#include "userdata.h"

using namespace netCDF;

void create_dir_path(const std::string& input) {
//     std::string dirname = input.substr(0, input.find_last_of('/'));
    std::string dirname = "mkdir -p $(dirname "+input+")";
    char* call = new char [dirname.size()+1];
//     std::cerr << dirname << std::endl;
    std::copy(dirname.begin(), dirname.end(), call);
    call[dirname.size()]='\0';

    std::system(call);
    delete[] call;
}


struct export_species_netcdf {
    export_species_netcdf(NcFile& _file, const std::string& _name, const std::vector<NcDim>& _axes, const NcDim& _time) : name(_name) {
        species = _file.addGroup(name);

        vn = species.addVar("vn", NC_DOUBLE, _axes[1]);
        vp = species.addVar("vp", NC_DOUBLE, _axes[2]);

        vn_T = species.addVar("vn_T", NC_DOUBLE);
        vp_T = species.addVar("vp_T", NC_DOUBLE);
        n0 = species.addVar("n0", NC_DOUBLE);
        m = species.addVar("m", NC_DOUBLE);
        q = species.addVar("q", NC_DOUBLE);

        f = species.addVar("f", NC_DOUBLE, _axes);
        E = species.addVar("E", NC_DOUBLE, _axes[0]);
        n = species.addVar("n", NC_DOUBLE, _axes[0]);
        lmbda = species.addVar("lambda", NC_DOUBLE, _axes[0]);

        n_evolution = species.addVar("n_evolution", NC_DOUBLE, _time);
        kinetic_energy_evolution = species.addVar("kinetic_energy_evolution", NC_DOUBLE, _time);
        momentum_flux_r_evolution = species.addVar("momentum_flux_r_evolution", NC_DOUBLE, _time);
        momentum_flux_l_evolution = species.addVar("momentum_flux_l_evolution", NC_DOUBLE, _time);
        heat_flux_evolution = species.addVar("heat_flux_evolution", NC_DOUBLE, _time);
        energy_flux_evolution = species.addVar("energy_flux_evolution", NC_DOUBLE, _time);
        temp_p_evolution = species.addVar("temperature_p_evolution", NC_DOUBLE, _time);
        temp_n_evolution = species.addVar("temperature_n_evolution", NC_DOUBLE, _time);

        description = species.addVar("description", NC_STRING);

        f.setCompression(false,true,9);
        n_evolution.setCompression(false,true,9);
        kinetic_energy_evolution.setCompression(false,true,9);
        momentum_flux_l_evolution.setCompression(false,true,9);
        momentum_flux_r_evolution.setCompression(false,true,9);
        heat_flux_evolution.setCompression(false,true,9);
        energy_flux_evolution.setCompression(false,true,9);
        temp_p_evolution.setCompression(false,true,9);
        temp_n_evolution.setCompression(false,true,9);
    }

    std::string name;
    NcGroup species;
    NcVar vn, vp, vn_T, vp_T, n0, m, q, f, E, n, lmbda, description, n_evolution, kinetic_energy_evolution, heat_flux_evolution, energy_flux_evolution, temp_n_evolution, temp_p_evolution, momentum_flux_r_evolution, momentum_flux_l_evolution;

    void write(const UserData& data) {
        std::vector<realtype> tmp_data;
        realtype tmp = 0.;

        tmp_data.resize(data.mx);

        tmp_data.resize(data.mvp);
        tmp=data.vpstart;
        for(size_t k=0; k<data.mvp; k++){
            tmp_data[k]=tmp;
            tmp+=data.dvp;
        }
        vp.putVar(tmp_data.data());

        tmp_data.resize(data.mvn);
        tmp=data.vnstart;
        for(size_t k=0; k<data.mvn; k++){
            tmp_data[k]=tmp;
            tmp+=data.dvn;
        }
        vn.putVar(tmp_data.data());

        vp_T.putVar(&data.vp_T_nominal);
        vn_T.putVar(&data.vn_T_nominal);
        n0.putVar(&data.n_nominal);
        lmbda.putVar(&lambda);
        m.putVar(&data.m);
        q.putVar(&data.q);

        f.putVar(N_VGetArrayPointer(data.f));
        n.putVar(data.n.data());
        E.putVar(N_VGetArrayPointer(*data.E));

        kinetic_energy_evolution.putVar(data.kinetic_energy.data());
        n_evolution.putVar(data.n_total.data());
        heat_flux_evolution.putVar(data.heat_flux.data());
        energy_flux_evolution.putVar(data.energy_flux.data());
        momentum_flux_r_evolution.putVar(data.momentum_flux_r.data());
        momentum_flux_l_evolution.putVar(data.momentum_flux_l.data());
        temp_p_evolution.putVar(data.temp_p.data());
        temp_n_evolution.putVar(data.temp_n.data());

        description.putVar(&data.description);
    }
};

template<typename ... DATA>
void export_to_netcdf(const std::string& outpath, const DATA& ... data) {
    NcDim time;
    std::vector<NcDim> axes(3);
    std::vector<export_species_netcdf> species {};

    NcVar electric_energy, x, v_normal, L, dt, E;
    std::vector<const UserData*> data_v = {&data ...};

    create_dir_path(outpath);
    NcFile file = NcFile(outpath, NcFile::replace);

    axes[0] = file.addDim("x", data_v[0]->mx);
    axes[1] = file.addDim("vn", data_v[0]->mvn);
    axes[2] = file.addDim("vp", data_v[0]->mvp);
    time = file.addDim("t",data_v[0]->electric_energy->size());

    electric_energy = file.addVar("electric_energy", NC_DOUBLE, time);

    x = file.addVar("x", NC_DOUBLE, axes[0]);
    v_normal = file.addVar("v_normalizaion", NC_DOUBLE);
    L = file.addVar("X_length", NC_DOUBLE);
    dt = file.addVar("dt", NC_DOUBLE);
    E = file.addVar("E", NC_DOUBLE, axes[0]);

    species.reserve(data_v.size());
    for(size_t i=0; i<data_v.size(); ++i){
        species.push_back(export_species_netcdf(file,data_v[i]->description,axes,time));
    }

    realtype tmp;
    std::vector<realtype> tmp_data;
    tmp_data.resize(data_v[0]->mx);
    tmp=data_v[0]->xstart;
    for(size_t k=0; k<data_v[0]->mx; k++){
        tmp_data[k]=tmp;
        tmp+=data_v[0]->dx;
    }
    x.putVar(tmp_data.data());

    v_normal.putVar(&v_normalization);
    L.putVar(&X_LENGTH);
    electric_energy.putVar(data_v[0]->electric_energy->data());

    E.putVar(N_VGetArrayPointer(*data_v[0]->E));
    dt.putVar(&DTOUT);

    for(size_t i=0; i<data_v.size(); ++i){
        species[i].write(*data_v[i]);
    }

    std::cerr << std::endl << "Exporting " << outpath << std::endl;
}
