#pragma once

#include <sundials/sundials_nvector.h>
#include <nvector/nvector_serial.h>    /* access to serial N_Vector            */
#include "sundial_stuff.h"

#include <iostream>

#include <omp.h>
#ifdef _OPENMP
#include <nvector/nvector_openmp.h>    /* serial N_Vector types, fcts., macros */
#endif


#ifdef _OPENMP
const size_t num_threads = omp_get_max_threads();
#else
const size_t num_threads = 1;
#endif


N_Vector new_nv( size_t nums ) {
    N_Vector output;
#ifdef _OPENMP
    output = N_VNew_OpenMP(nums, num_threads);
    const char* funcname = "N_VNew_OpenMP";
#else
    output = N_VNew_Serial(nums);
    const char* funcname = "N_VNew_Serial";
#endif
    if (check_flag( (void*) output, funcname, 0)) exit(1);
    return output;
}

void destroy_nv( N_Vector& nv ) {
#ifdef _OPENMP
    N_VDestroy_OpenMP(nv);
#else
    N_VDestroy_Serial(nv);
#endif
}

void omp_message(){
#ifdef _OPENMP
    std::cerr << "Running with " << num_threads << " omp threads." << std::endl;
#endif
}
