#pragma once

/*
 *-------------------------------
 * Functions called by the solver
 *-------------------------------
 */

/* f routine. Compute f(t,u). */

extern measure_time<double> mt_f;

#include "boundary.h"

template<size_t, btype = btype::PERIODIC>
constexpr realtype x_advection(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp, const realtype& maxwellian_n = 0.);

template<>
constexpr realtype x_advection<1, btype::PERIODIC>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp, const realtype&) {

    realtype u_tmp_x=0.;
    if (vp<0) {
        u_tmp_x = IJKth(udata,(iter_x+1)%MX,iter_vn,iter_vp);
        return -vp*(u_tmp_x-uijk)*data->rezi_dx;
    } else {
        u_tmp_x = IJKth(udata,(iter_x-1+MX)%MX,iter_vn,iter_vp);
        return -vp*(uijk-u_tmp_x)*data->rezi_dx;
    }
}

template<>
constexpr realtype x_advection<1, btype::DIRICHLET>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp, const realtype& maxwellian_n) {

    realtype u_tmp_x=0.;
    if (vp<0) {
        if (iter_x == MX-1) {
            u_tmp_x = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
        } else {
            u_tmp_x = IJKth(udata,iter_x+1,iter_vn,iter_vp);
        }
        return -vp*(u_tmp_x-uijk)*data->rezi_dx;
    } else {
        if (iter_x == 0) {
            u_tmp_x = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
        } else {
            u_tmp_x = IJKth(udata,iter_x-1,iter_vn,iter_vp);
        }
        return -vp*(uijk-u_tmp_x)*data->rezi_dx;
    }
}

template<>
constexpr realtype x_advection<2, btype::PERIODIC>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp, const realtype&) {

    realtype u_tmp_x=0.;
    realtype u_tmp_x2=0.;
    if (vp<0) {
        u_tmp_x = IJKth(udata,(iter_x+1)%MX,iter_vn,iter_vp);
        u_tmp_x2 = IJKth(udata,(iter_x+2)%MX,iter_vn,iter_vp);
        return -vp*(-u_tmp_x2+4.*u_tmp_x-3.*uijk)*.5*data->rezi_dx;
    } else {
        u_tmp_x = IJKth(udata,(iter_x-1+MX)%MX,iter_vn,iter_vp);
        u_tmp_x2 = IJKth(udata,(iter_x-2+MX)%MX,iter_vn,iter_vp);
        return -vp*(3.*uijk-4.*u_tmp_x+u_tmp_x2)*.5*data->rezi_dx;
    }
}

template<>
constexpr realtype x_advection<2, btype::DIRICHLET>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp, const realtype& maxwellian_n) {

    realtype u_tmp_x=0.;
    realtype u_tmp_x2=0.;
    if (vp<0) {
        switch (iter_x) {
            case MX-1:
                u_tmp_x = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                u_tmp_x2 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                break;
            case MX-2:
                u_tmp_x = IJKth(udata,MX-1,iter_vn,iter_vp);
                u_tmp_x2 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                break;
            default:
                u_tmp_x = IJKth(udata,iter_x+1,iter_vn,iter_vp);
                u_tmp_x2 = IJKth(udata,iter_x+2,iter_vn,iter_vp);
        }
        return -vp*(-u_tmp_x2+4.*u_tmp_x-3.*uijk)*.5*data->rezi_dx;
    } else {
        switch (iter_x) {
            case 1:
                u_tmp_x = IJKth(udata,0,iter_vn,iter_vp);
                u_tmp_x2 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                break;
            case 0:
                u_tmp_x = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                u_tmp_x2 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                break;
            default:
                u_tmp_x = IJKth(udata,iter_x-1,iter_vn,iter_vp);
                u_tmp_x2 = IJKth(udata,iter_x-2,iter_vn,iter_vp);
        }
        return -vp*(3.*uijk-4.*u_tmp_x+u_tmp_x2)*.5*data->rezi_dx;
    }
}

template<>
constexpr realtype x_advection<3, btype::PERIODIC>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp, const realtype&) {

    realtype u_tmp_x_1=0.;
    realtype u_tmp_x=0.;
    realtype u_tmp_x2=0.;
    if (vp<0) {
        u_tmp_x_1 = IJKth(udata,(iter_x-1+MX)%MX,iter_vn,iter_vp);
        u_tmp_x = IJKth(udata,(iter_x+1)%MX,iter_vn,iter_vp);
        u_tmp_x2 = IJKth(udata,(iter_x+2)%MX,iter_vn,iter_vp);
        return -vp*(-u_tmp_x2+6.*u_tmp_x-3.*uijk-2.*u_tmp_x_1)/6.*data->rezi_dx;
    } else {
        u_tmp_x_1 = IJKth(udata,(iter_x+1)%MX,iter_vn,iter_vp);
        u_tmp_x = IJKth(udata,(iter_x-1+MX)%MX,iter_vn,iter_vp);
        u_tmp_x2 = IJKth(udata,(iter_x-2+MX)%MX,iter_vn,iter_vp);
        return -vp*(3.*uijk-6.*u_tmp_x+u_tmp_x2+2.*u_tmp_x_1)/6.*data->rezi_dx;
    }
}

template<>
constexpr realtype x_advection<3, btype::DIRICHLET>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp, const realtype& maxwellian_n) {

    realtype u_tmp_x_1=0.;
    realtype u_tmp_x=0.;
    realtype u_tmp_x2=0.;
    if (vp<0) {
        switch (iter_x) {
            case MX-1:
                u_tmp_x_1 = IJKth(udata,MX-2,iter_vn,iter_vp);
                u_tmp_x = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                u_tmp_x2 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                break;
            case MX-2:
                u_tmp_x_1 = IJKth(udata,MX-3,iter_vn,iter_vp);
                u_tmp_x = IJKth(udata,MX-1,iter_vn,iter_vp);
                u_tmp_x2 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                break;
            case 0:
                u_tmp_x_1 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                u_tmp_x = IJKth(udata,1,iter_vn,iter_vp);
                u_tmp_x2 = IJKth(udata,2,iter_vn,iter_vp);
                break;
            default:
                u_tmp_x_1 = IJKth(udata,iter_x-1,iter_vn,iter_vp);
                u_tmp_x = IJKth(udata,iter_x+1,iter_vn,iter_vp);
                u_tmp_x2 = IJKth(udata,iter_x+2,iter_vn,iter_vp);
        }
        return -vp*(-u_tmp_x2+6.*u_tmp_x-3.*uijk-2.*u_tmp_x_1)/6.*data->rezi_dx;
    } else {
        switch (iter_x) {
            case 1:
                u_tmp_x_1 = IJKth(udata,2,iter_vn,iter_vp);
                u_tmp_x = IJKth(udata,0,iter_vn,iter_vp);
                u_tmp_x2 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                break;
            case 0:
                u_tmp_x_1 = IJKth(udata,1,iter_vn,iter_vp);
                u_tmp_x = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                u_tmp_x2 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                break;
            case MX-1:
                u_tmp_x_1 = maxwellian(vp,data->vp_T_nominal)*maxwellian_n;
                u_tmp_x = IJKth(udata,MX-2,iter_vn,iter_vp);
                u_tmp_x2 = IJKth(udata,MX-3,iter_vn,iter_vp);
                break;
            default:
                u_tmp_x_1 = IJKth(udata,iter_x+1,iter_vn,iter_vp);
                u_tmp_x = IJKth(udata,iter_x-1,iter_vn,iter_vp);
                u_tmp_x2 = IJKth(udata,iter_x-2,iter_vn,iter_vp);
        }
        return -vp*(3.*uijk-6.*u_tmp_x+u_tmp_x2+2.*u_tmp_x_1)/6.*data->rezi_dx;
    }
}



template<size_t, btype = btype::PERIODIC>
constexpr realtype vp_advection(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp_adv_coeffi);

template<>
constexpr realtype vp_advection<1, btype::PERIODIC>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp_adv_coeffi) {
    realtype u_tmp_vp=0.;

    if (vp_adv_coeffi<0.) {
        u_tmp_vp = IJKth(udata,iter_x,iter_vn,(iter_vp+1)%MVp);
        return -vp_adv_coeffi * (u_tmp_vp - uijk)*data->rezi_dvp;
    } else {
        u_tmp_vp = IJKth(udata,iter_x,iter_vn,(iter_vp-1+MVp)%MVp);
        return -vp_adv_coeffi * (uijk - u_tmp_vp)*data->rezi_dvp;
    }
}

template<>
constexpr realtype vp_advection<2, btype::PERIODIC>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp_adv_coeffi) {
    realtype u_tmp_vp=0.;
    realtype u_tmp_vp2=0.;

    if (vp_adv_coeffi<0.) {
        u_tmp_vp = IJKth(udata,iter_x,iter_vn,(iter_vp+1)%MVp);
        u_tmp_vp2 = IJKth(udata,iter_x,iter_vn,(iter_vp+2)%MVp);
        return -vp_adv_coeffi*(-u_tmp_vp2+4.*u_tmp_vp-3.*uijk)*.5*data->rezi_dvp;
    } else {
        u_tmp_vp = IJKth(udata,iter_x,iter_vn,(iter_vp-1+MVp)%MVp);
        u_tmp_vp2 = IJKth(udata,iter_x,iter_vn,(iter_vp-2+MVp)%MVp);
        return -vp_adv_coeffi*(3.*uijk-4.*u_tmp_vp+u_tmp_vp2)*.5*data->rezi_dvp;
    }
}

template<>
constexpr realtype vp_advection<3, btype::PERIODIC>(realtype* udata, const UserData* data, const size_t& iter_x, const size_t& iter_vn, const size_t& iter_vp, const realtype& uijk, const realtype& vp_adv_coeffi) {
    realtype u_tmp_vp_1=0.;
    realtype u_tmp_vp=0.;
    realtype u_tmp_vp2=0.;

    if (vp_adv_coeffi<0.) {
        u_tmp_vp_1 = IJKth(udata,iter_x,iter_vn,(iter_vp+(-1+MVp))%MVp);
        u_tmp_vp = IJKth(udata,iter_x,iter_vn,(iter_vp+1)%MVp);
        u_tmp_vp2 = IJKth(udata,iter_x,iter_vn,(iter_vp+2)%MVp);
        return -vp_adv_coeffi*(-u_tmp_vp2+6.*u_tmp_vp-3.*uijk-2.*u_tmp_vp_1)/6.*data->rezi_dvp;
    } else {
        u_tmp_vp_1 = IJKth(udata,iter_x,iter_vn,(iter_vp+1)%MVp);
        u_tmp_vp = IJKth(udata,iter_x,iter_vn,(iter_vp+(MVp-1))%MVp);
        u_tmp_vp2 = IJKth(udata,iter_x,iter_vn,(iter_vp+(-2+MVp))%MVp);
        return -vp_adv_coeffi*(3.*uijk-6.*u_tmp_vp+u_tmp_vp2+2.*u_tmp_vp_1)/6.*data->rezi_dvp;
    }
}

template<class ... c_operators>
struct add_colls;
template<class c_operator, class ... c_operators>
struct add_colls<c_operator, c_operators...>{
    static realtype collide(const realtype& _is,
                            const UserData& _userdata,
                            const size_t& _iter_x,
                            const realtype& _vp,
                            const realtype& _vn
    ) {
        return c_operator::OP(_is,_userdata,_iter_x,_vp,_vn) + add_colls<c_operators...>::collide(_is,_userdata,_iter_x,_vp,_vn);
    }
};
template<>
struct add_colls<>{
    static realtype collide(const realtype&,
                            const UserData&,
                            const size_t&,
                            const realtype&,
                            const realtype&) {
        return 0.;
    }
};

template<class ... c_operators>
static int f_collisionless(realtype /*t*/, N_Vector u, N_Vector udot, void *user_data) {
    static_assert(boundary_x==btype::PERIODIC || boundary_x==btype::DIRICHLET);

    realtype* udata = N_VGetArrayPointer(u);
    realtype* dudata = N_VGetArrayPointer(udot);
    UserData* data = (UserData*) user_data;
    realtype* Edata = N_VGetArrayPointer(*(data->E));

//     std::cerr << "Here: "<< data->description << std::endl;
    mt_f.start();
#pragma omp parallel for
    for ( size_t iter_x=0; iter_x < MX; iter_x++) {
        realtype vp_adv_coeffi = Edata[iter_x] * data->advection_coeffi;
        realtype boundary_value_x = 0.;
        realtype boundary_value_vn = 0., boundary_value_vp = 0.;

        realtype vn = data->vnstart;
        for ( size_t iter_vn=0; iter_vn < MVn; iter_vn++, vn += data->dvn) {
            if constexpr (boundary_x == btype::DIRICHLET) {
                if (iter_x == 0 || iter_x == MX-1)
                    boundary_value_vn = maxwellian(vn,data->vn_T_nominal)*boundary_value_x;
            }

            realtype vp = data->vpstart;
            for ( size_t iter_vp=0; iter_vp < MVp; iter_vp++, vp+=data->dvp) {
                if constexpr (boundary_x == btype::DIRICHLET) {
                    if (iter_x == 0 || iter_x == MX-1)
                        boundary_value_vp = maxwellian(vp,data->vp_T_nominal)*boundary_value_vn;
                }

                realtype uijk = IJKth(udata, iter_x, iter_vn, iter_vp ); // center
                static_assert(boundary_x == btype::DIRICHLET);
                IJKth(dudata, iter_x, iter_vn, iter_vp) =
                    + x_advection<2, boundary_x>(udata, data, iter_x, iter_vn, iter_vp, uijk, vp, boundary_value_vp)
                    + vp_advection<order, btype::PERIODIC>(udata, data, iter_x, iter_vn, iter_vp, uijk, vp_adv_coeffi)
                    + add_colls<c_operators...>::collide(uijk,*data,iter_x,vp,vn) //TODO: can't this be done directly with an expansion loci?
                    ;
            }
        }
//         exit(1);
    }
    mt_f.stop();
    return(0);
}
